include std/text.e
include std/io.e
include std/filesys.e
include std/sequence.e
include std/search.e
include std/error.e
include myLibs/myDebug.e
include myLibs/common.e
include tin/tinEWG.ew
include tin/kisTinEWG.e

constant
    App     = "Code Beautifier",
    Ver     = "v0.1.0"
    
constant BLOCK_TAGS = {
  {"if","elsif","else","end if"},
  {"ifdef","elsifdef","elsedef","end ifdef"},
  {"switch", "case", "end switch"},
  {"while","entry","end while"},
  {"loop","entry","end loop"},
  {"for","end for"},
  {"type","end type"},
  {"procedure","end procedure"},
  {"function","end function"}
}

constant LINE=1, WORD=2, COL=3

integer eolLength = length(EOLSEP)

sequence InitialDir = {}

sequence allWords = {}
sequence startTags = {}, midTags = {}
integer f_out

integer
  Form1, btnOpen, mleResult

------------------------------------------------------------------------------

procedure extractTags()
  for i = 1 to length(BLOCK_TAGS) do
    startTags = append(startTags, BLOCK_TAGS[i][1])
    midTags = append(midTags, BLOCK_TAGS[i][2..$-1])
  end for
end procedure

------------------------------------------------------------------------------

function splitAny(sequence source, object delim=", \t|", integer no_empty=0)
  sequence ret = {}
  integer start = 1, pos, next_pos, str

  if length(delim) = 0 then
    return {source}
  end if

  while 1 do
    integer strFrom = find('"', source, start)
    integer strTo = find('"', source, strFrom+1)
    pos = search:find_any(delim, source, start)
    if (pos > strFrom) and (pos < strTo) then
      start = strTo+1
      pos = search:find_any(delim, source, start)
    end if
    next_pos = pos + 1
    if pos then
      ret = append(ret, source[start..pos-1])
      start = next_pos
    else
      exit
    end if
  end while

  ret = append(ret, source[start..$])

  integer k = length(ret)
  if no_empty then
    k = 0
    for i = 1 to length(ret) do
      if length(ret[i]) != 0 then
        k += 1
        if k != i then
          ret[k] = ret[i]
        end if
      end if
    end for
  end if

  if k < length(ret) then
    return ret[1 .. k]
  else
    return ret
  end if
end function

------------------------------------------------------------------------------

function getWords(sequence lines)
  sequence words = {}
  integer p
  for l = 1 to length(lines) do
    integer start = 1
    sequence s = splitAny(lines[l], " ({[]}),;&+-=\t")
    integer i = 1
    while i <= length(s) do
      if length(s[i]) then
        puts(f_debug, s[i] & "\n")
        flush(f_debug)
        p = match(s[i], lines[l], start)
        if equal(s[i], {34}) then  -- double quote
          start = match({34}, lines[l], p+1)
          i = find({34}, s, i+1)
        else
          words = append(words, {l, s[i], p})
          start = p+1
        end if
      end if
      i += 1
    end while
  end for
  return words
end function

------------------------------------------------------------------------------

procedure beautify(sequence lines)
  integer offset = 0, nextOffset = 0
--  integer withinComment = 0  -- full lines between /* and */
  extractTags()
  analyzeObject(startTags, "startTags", f_debug)
  analyzeObject(midTags, "midTags", f_debug)
  for l = 1 to length(lines) do
    analyzeObject(lines[l], sprintf("lines[%d]",l), f_debug)
    sequence wordsFound = commonField(allWords, LINE, l)
    analyzeObject(wordsFound, "wordsFound", f_debug)
    integer sc = match("--", lines[l])
--    integer mcs = match("/*", lines[l])
--    integer mce = match("*/", lines[l])

    sequence words = {}
    for i = 1 to length(wordsFound) do
      if (sc = 0) or (allWords[wordsFound[i]][COL] < sc) then
        words = append(words, allWords[wordsFound[i]][WORD])
      end if
    end for
    analyzeObject(words, "words", f_debug)

    nextOffset = -1
    for i = 1 to length(BLOCK_TAGS) do
      sequence matches = find_all(startTags[i], words)
      switch length(matches) do
        case 0 then
          for j = 1 to length(midTags[i]) do
            matches = find_all(midTags[i][j], words)
            analyzeObject(matches, "matches", f_debug)
            switch length(matches) do
              case 0 then
                -- offset unchanged
              case else
                nextOffset = offset  
                offset -= 2
            end switch
          end for
        case 1 then
          analyzeObject(words[matches[1]], sprintf("words[%d]", matches[1]), f_debug)
          if (matches[1] > 1) and equal(words[matches[1]-1], "end") then
            offset -= 2
          else
            nextOffset = offset+2  
          end if
        case else
          -- offset unchanged
      end switch
        
    end for
    printf(f_debug, "offset = %d, next offset = %d\n", {offset, nextOffset})
    flush(f_debug)
    sequence s = repeat(' ', offset) & trim(lines[l])
    puts(f_out, s & "\n")  
    flush(f_out)
    AppendText(getHandle(mleResult), s & "\r\n")
    if nextOffset != -1 then offset = nextOffset end if
  end for
end procedure

------------------------------------------------------------------------------

procedure onClick_Open()
  sequence fName = FileDlg( Open, InitialDir&SLASH&"*.e*",
                            "Euphoria files (*.e*)|*.e*" )
  if length(fName) then
    f_out = open(fName&".btf", "w")
    sequence lines = read_lines(fName)
    allWords = getWords(lines)
--    analyzeObject(allWords, "allWords", f_debug)
    beautify(lines)
    close(f_out)
  end if
end procedure

------------------------------------------------------------------------------

-- Application Initialization
function appInit()
  integer lRC

  sequence cmd = command_line()
  sequence info = pathinfo(cmd[2])
  void = chdir(info[PATH_DIR])
  InitialDir = current_dir()
  sequence cur_date = date()
  f_debug = open(InitialDir&SLASH&"debug.log", "w")
  crash_file(InitialDir&SLASH&"ex.err")
  debugLevel = 1

  lRC = 0

  -- The window
  Form1 = create( Form, App & " " & Ver, 1, 1, 1024, 726 )

  btnOpen = create( Button, "Open", 8, 4, 80, 21 )
  mleResult = create( MultiEdit, "", 8, 28, 992, 650 )

  atom courierNew = NewFont("Courier New", 10, False, False, False)
  SetFont( getHandle(mleResult), courierNew)

  setHandler(btnOpen, Click, routine_id( "onClick_Open" ))

  return lRC
end function   -- appInit

------------------------------------------------------------------------------

-- Main
if appInit() = 0 then
  setText(mleResult, "")
  eventLoop()
  close(f_debug)
  logMsg("Window closes")
  --Close the application and free memory
  CloseApp(0)
end if
