-- kisTinEWG.e
-- Keep it Simple tinEWG
-- Jean-Marc Duro
-- 2015-11-28
-- v0.0.3

include std/search.e
include std/filesys.e
include std/convert.e
include tin/tinEWG.ew as tin
include myLibs/myDebug.e

-- stores object ids with their class
export sequence registry = {}
export sequence myControls = {}

public constant REG_HWND=1, REG_CLASS=2, REG_CHILDREN=3, REG_PARAMS=4, REG_TIP=5
public constant TIP_NONE=0, TIP_AUTO=1
public constant Form = 100, SimCombo = 101

constant NEW_WIDGETS = {Form, SimCombo}

constant TINEWG_CLASSES = {
  {Button, "Button"},
--  {Calendar, "Calendar"},
  {Check, "Check"},
  {ClickImage, "ClickImage"},
  {ClickLabel, "ClickLabel"},
  {ClickPicture, "ClickPicture"},
  {ClickText, "ClickText"},
  {ComboBox, "ComboBox"},
  {Dialog, "Dialog"},
  {DropDown, "DropDown"},
  {Edit, "Edit"},
--  {Fake, "Fake"},
  {Form, "Form"},
  {FramedText, "FramedText"},
  {Group, "Group"},
  {IconButton, "IconButton"},
  {Image, "Image"},
--  {IpEdit, "IpEdit"},
  {Label, "Label"},
  {List, "List"},
  {ListView, "ListView"},
--  {MainWin, "MainWin"},
  {MultiEdit, "MultiEdit"},
  {Picture, "Picture"},
  {PictureButton, "PictureButton"},
  {PicturePushButton, "PicturePushButton"},
  {ProgressBar, "ProgressBar"},
  {PushButton, "PushButton"},
  {Radio, "Radio"},
  {RichEdit, "RichEdit"},
  {SelecList, "SelecList"},
  {SimCheck, "SimCheck"},
  {SimCombo, "SimCombo"},
  {SimEdit, "SimEdit"},
  {SimMultiEdit, "SimMultiEdit"},
  {SimMultiList, "SimMultiList"},
  {SimRadio, "SimRadio"},
  {SortedList, "SortedList"},
  {SortedSelecList, "SortedSelecList"},
  {StatusBar, "StatusBar"},
  {Text, "Text"},
  {ToolTip, "ToolTip"} --,
--  {UserDefControl, "UserDefControl"}
}

public sequence tinewgDir = "tin"&SLASH
sequence resourceDir = tinewgDir&"res"&SLASH

------------------------------------------------------------------------------

-- get the handle of a widget
public function getHandle(integer id)
  return registry[id][REG_HWND]
end function

------------------------------------------------------------------------------

-- get the class of a widget
public function getClass(integer id)
  return registry[id][REG_CLASS]
end function

------------------------------------------------------------------------------

-- get the children of a widget
public function getChildren(integer id)
  return registry[id][REG_CHILDREN]
end function

------------------------------------------------------------------------------

-- get the parameters of a widget
public function getParams(integer id)
  return registry[id][REG_PARAMS]
end function

------------------------------------------------------------------------------

-- set the children of a widget
public procedure setChildren(integer id, sequence children)
  registry[id][REG_CHILDREN] = children
end procedure

------------------------------------------------------------------------------

-- set the parameters of a widget
public procedure setParams(integer id, sequence parameters)
  registry[id][REG_PARAMS] = parameters
end procedure

------------------------------------------------------------------------------

-- find a Widget ID with its handle or the handle of its child
public function findWidgetId(integer hwnd)
  if hwnd = 0 then return 0 end if
  sequence s = find_nested(hwnd, registry)
--  analyzeObject(s, "s", f_debug)
  if length(s) then
    return s[1]
  else
    return 0
  end if
end function

------------------------------------------------------------------------------

-- get the class name of a widget
public function getClassName(integer id)
  object class = getClass(id)
  return vlookup(class, TINEWG_CLASSES, 1, 2, "Unknown")
end function

------------------------------------------------------------------------------

-- create a widget and store it in registry
public function create( integer class, sequence text="", integer x=0, integer y=0,
                        integer w=0, integer h=0, sequence params={}, object tip=TIP_NONE )
  atom id
  sequence children = {}, parms = {}

  if atom(tip) and (tip=TIP_AUTO) then
    tip = vlookup(class, TINEWG_CLASSES, 1, 2, "Unknown")
  end if

  if find(class, NEW_WIDGETS) then
    switch class do
    case Form then
      integer oldType = WindowType
      if length(params) = 0 then
        WindowType = StandardWin
      else
        WindowType = params[1]
      end if
      Window( text, x, y, w, h )
      WindowType = oldType
      id = WinHwnd

    case SimCombo then
      id = 0 - ( length(myControls)+1 )
      sequence items
      integer visibleItems
      if length(params >= 2) then
        items = params[1]
        visibleItems = params[2]
      else
        items = {}
        visibleItems = 0
      end if
      children = {
        Control(PictureButton,"",x+w-18,y,17,h),
        Control(Edit,text,x,y,w-17,h),
        Control(List,"",x,y+21,w,h*visibleItems)
      }
      myControls = append(myControls, {id, children})
      for i = 1 to length(items) do
        if atom(items[i]) then
          ListAdd(children[3],to_string(items[i]))
        else
          ListAdd(children[3],items[i])
        end if
      end for
      ShowWindow (children[3],SW_HIDE)
      SetPicture(children[1], PictureButton, resourceDir&"not_dropped.bmp")
      parms = {0}  -- not dropped
      if sequence(tip) then
        for i = 1 to length(children) do
          AddToolTip(children[i], tip)
        end for
      end if

    end switch

  elsif (class = DropDown) or (class = ComboBox) then
    id = Control( class, text, x, y, w, h )
    if sequence(tip) then
      AddToolTip(id, tip)
    end if
    sequence items
    integer visibleItems
    if length(params >= 2) then
      items = params[1]
      visibleItems = params[2]
    else
      items = {}
      visibleItems = 0
    end if
    ListAdd(id, items)

  elsif class = Group then
    id = Control( class, text, x, y, w, h )
    children = params
    if atom(children) then
      SetGroup(children, id)
    else
      for i = 1 to length(children) do
        SetGroup(children[i], id)
      end for
    end if

  elsif class = StatusBar then
    id = Control( class, text, x, y, w, h )
    if length(params) then
      sb_SetParts(id, params)
    end if

  elsif (class = Button) or (class = PictureButton) then
    id = Control( class, text, x, y, w, h )
    if length(params) then
      SetEnable(id, params[1])
    end if

  else  
    id = Control( class, text, x, y, w, h )
    if sequence(tip) then
      AddToolTip(id, tip)
    end if

  end if

  registry = append(registry, {id, class, children, parms, tip})
  return length(registry)
end function

------------------------------------------------------------------------------

-- set the event handler of a widget
public procedure setHandler(integer id, integer event, integer event_routine=0)
  atom class = getClass(id)
  atom hwnd = getHandle(id)
  if find(class, NEW_WIDGETS) then
    switch class do
    end switch
  else
    SetHandler(hwnd, event, event_routine)
  end if
end procedure

------------------------------------------------------------------------------

-- manage events
public procedure manageEvents(object id)
  if atom(id) then id = {id} end if
  analyzeObject(id, "id", f_debug)
  for i = 1 to length(id) do
    atom class = getClass(id[i])
    analyzeObject(class, "class", f_debug)
    if find(class, NEW_WIDGETS) then
      switch class do
      case SimCombo then
        sequence children = getChildren(id[i])
        analyzeObject(children, "children", f_debug)
        atom pictureButton = children[1]
        atom editText = children[2]
        atom listItems = children[3]
        sequence params = getParams(id[i])
        integer dropped = params[1]

        if EventOwner = pictureButton then
          if not dropped then
            SetPicture(pictureButton,tin:PictureButton,resourceDir&"dropped_down.bmp")
            setParams(id[i], {1})  -- set dropped
            ShowWindow (listItems,SW_SHOW)
          else
            SetPicture(pictureButton,tin:PictureButton,resourceDir&"not_dropped.bmp")
            setParams(id[i], {0})  -- set not dropped
            ShowWindow (listItems,SW_HIDE)
            SetFocus(editText)
          end if
        elsif EventOwner = editText then
          --Note: when the message disappears the control looses the keyboard input and must be reactivated
          ActivateText(editText)
        elsif EventOwner = listItems then
          SetText(editText, GetItem(listItems))
          SetPicture(pictureButton,tin:PictureButton,resourceDir&"not_dropped.bmp")
          setParams(id[i], {0})  -- set not dropped
          ShowWindow (listItems,SW_HIDE)
          SetFocus(editText)
        end if
      end switch
    else
    end if
  end for
end procedure

------------------------------------------------------------------------------

--The Event Loop
public procedure eventLoop(integer rtn_id=0)
  --Run continuosly the event checking
  while tin:True do
    --Wait for an event
    WaitEvent()
    if UserTrap then
      analyzeObject(UserTrap, "UserTrap", f_debug)
    end if    
    if UserTrap and rtn_id then
      --Retrieve the message data (e.g. hwnd,message,wParam,lParam) and store it
      --into our custom sequence wpdata
      sequence wpdata = GetTrappedMessage()
      analyzeObject(wpdata, "wpdata", f_debug)
      call_proc(rtn_id, wpdata)
    end if    
    integer id = findWidgetId(EventOwner)
    if id = 0 then continue end if
    atom class = getClass(id)
    --The event was a left mouse click?
    if Event = Click then
      analyzeObject(EventOwner, "EventOwner", f_debug)
      analyzeObject(id, "id", f_debug)
      analyzeObject(class, "class", f_debug)
      if find(class, NEW_WIDGETS) then
        switch class do
        case SimCombo then
          manageEvents(id)
        end switch
      else
      end if
    --Close the application?
    elsif Event = Close then
 
      if class = Form then
        CloseApp(0)
      end if
 
    end if
    ProcessHandlers()
  end while
end procedure

------------------------------------------------------------------------------

-- set the text of a widget
public procedure setText(integer id, sequence text, integer section=0)
  atom class = getClass(id)
  atom hwnd = getHandle(id)
  if find(class, NEW_WIDGETS) then
    switch class do
    case SimCombo then
      sequence children = getChildren(id)
      atom editText = children[2]
      SetText(editText, text)
    end switch
  else
    SetText(hwnd, text, section)
  end if
end procedure

------------------------------------------------------------------------------

-- get the text of a widget
public function getText(integer id, integer section=0)
  atom class = getClass(id)
  atom hwnd = getHandle(id)
  if find(class, NEW_WIDGETS) then
    switch class do
    case SimCombo then
      sequence children = getChildren(id)
      atom editText = children[2]
      return GetText(editText)
    end switch
  else
    return GetText(hwnd)
  end if
  return 0
end function

------------------------------------------------------------------------------

-- set the enabled state of a widget
public procedure setEnable(integer id, integer state)
  atom class = getClass(id)
  atom hwnd = getHandle(id)
  if find(class, NEW_WIDGETS) then
    switch class do
    case SimCombo then
      sequence children = getChildren(id)
      for i = 1 to length(children) do
        SetEnable(children[i], state)
      end for
    end switch
  else
    SetEnable(hwnd, state)
  end if
end procedure

------------------------------------------------------------------------------

-- set the enabled state of a widget
public procedure setFocus(integer id)
  atom class = getClass(id)
  atom hwnd = getHandle(id)
  if find(class, NEW_WIDGETS) then
    switch class do
    case SimCombo then
      sequence children = getChildren(id)
      atom editText = children[2]
      SetFocus(editText)
    end switch
  else
    SetFocus(hwnd)
  end if
end procedure

